#include "library.h"

#include "eval.h"
#include "primitive.h"
#include "standard.h"
#include "macro.h"
#include "program.h"
#include "extension.h"

struct scm_thing * (* null_macros[])(struct scm_thing *, struct scm_thing *, struct scm_thing *) = {
  scm_quote,
  scm_lambda,
  scm_if,
  scm_set,
  scm_syntax_rules
};

const char * null_macro_names[] = {
  "quote",
  "lambda",
  "if",
  "set!",
  "syntax-rules"
};

struct scm_thing * null_variables;

struct scm_thing * (* interactive_macros[])(struct scm_thing *, struct scm_thing *, struct scm_thing *) = {
  scm_begin,
  scm_apply
};

const char * interactive_macro_names[] = {
  "begin",
  "apply"
};

struct scm_thing * (* interactive_procedures[])(struct scm_thing *) = {
  /* 6.1. Equivalence predicates */
  scm_eqv_p,
  scm_eq_p,
  scm_equal_p,
  /* 6.2.5. Numerical operations */
  scm_number_p,
  scm_complex_p,
  scm_real_p,
  scm_rational_p,
  scm_integer_p,
  scm_exact_p,
  scm_inexact_p,
  scm_numerical_equal,
  scm_numerical_less_than,
  scm_numerical_greater_than_or_equal_to,
  scm_numerical_greater_than,
  scm_numerical_less_than_or_equal_to,
  scm_zero_p,
  scm_positive_p,
  scm_negative_p,
  scm_odd_p,
  scm_even_p,
  scm_add,
  scm_subtract,
  scm_multiply,
  scm_divide,
  scm_abs,
  scm_quotient,
  scm_remainder,
  scm_modulo,
  scm_numerator,
  scm_denominator,
  scm_floor,
  scm_ceiling,
  scm_truncate,
  scm_round,
  /* 6.3. Other data types */
  /* 6.3.1. Booleans */
  scm_not,
  scm_boolean_p,
  /* 6.3.2. Pairs and lists */
  scm_pair_p,
  scm_cons,
  scm_car,
  scm_cdr,
  scm_set_car,
  scm_set_cdr,
  scm_null_p,
  scm_list_p,
  scm_list,
  scm_length,
  scm_append,
  scm_list_tail,
  scm_memq,
  scm_memv,
  scm_member,
  scm_assq,
  scm_assv,
  scm_assoc,
  /* 6.3.3. Symbols */
  scm_symbol_p,
  scm_symbol_to_string,
  scm_string_to_symbol,
  /* 6.3.4. Characters */
  scm_char_p,
  scm_char_equals_p,
  scm_char_less_than_p,
  scm_char_greater_than_p,
  scm_char_less_than_equals_p,
  scm_char_greater_than_equals_p,
  scm_char_ci_equals_p,
  scm_char_to_integer,
  scm_integer_to_char,
  /* 6.3.5. Strings */
  scm_string_p,
  scm_make_string,
  scm_string_length,
  scm_string_ref,
  scm_string_set,
  /* 6.3.6. Vectors */
  scm_vector_p,
  scm_make_vector,
  scm_vector_length,
  scm_vector_ref,
  scm_vector_set,
  scm_vector_fill,
  /* 6.4. Control features */
  scm_procedure_p,
  /* 6.5. Eval */
  scm_eval,
  /* 6.6. Input and output */
  /* 6.6.1. Ports */
  scm_input_port_p,
  scm_output_port_p,
  scm_current_input_port,
  scm_current_output_port,
  scm_open_input_file,
  scm_open_output_file,
  scm_close_input_port,
  scm_close_output_port,
  /* 6.6.2. Input */
  scm_read_char,
  scm_peek_char,
  scm_eof_object_p,
  /* 6.6.3. Output */
  scm_write_char,
  /* 6.6.4. System interface */
  /* scm_load, */
  /* my extensions */
  scm_snarf_file
};

const char * interactive_procedure_names[] = {
  /* 6.1. Equivalence predicates */
  "eqv?",
  "eq?",
  "equal?",
  /* 6.2.5. Numerical operations */
  "number?",
  "complex?",
  "read?",
  "rational?",
  "integer?",
  "exact?",
  "inexact?",
  "=",
  "<",
  ">=",
  ">",
  "<=",
  "zero?",
  "positive?",
  "negative?",
  "odd?",
  "even?",
  "+",
  "-",
  "*",
  "/",
  "abs",
  "quotient",
  "remainder",
  "modulo",
  "numerator",
  "denominator",
  "floor",
  "ceiling",
  "truncate",
  "round",
  /* 6.3. Other data types */
  /* 6.3.1. Booleans */
  "not",
  "boolean?",
  /* 6.3.2. Pairs and lists */
  "pair?",
  "cons",
  "car",
  "cdr",
  "set-car!",
  "set-cdr!",
  "null?",
  "list?",
  "list",
  "length",
  "append",
  "list-tail",
  "memq",
  "memv",
  "member",
  "assq",
  "assv",
  "assoc",
  /* 6.3.3. Symbols */
  "symbol?",
  "symbol->string",
  "string->symbol",
  /* 6.3.4. Characters */
  "char?",
  "char=?",
  "char<?",
  "char>?",
  "char<=?",
  "char>=?",
  "char-ci=?",
  "char->integer",
  "integer->char",
  /* 6.3.5. Strings */
  "string?",
  "make-string",
  "string-length",
  "string-ref",
  "string-set!",
  /* 6.3.6. Vectors */
  "vector?",
  "make-vector",
  "vector-length",
  "vector-ref",
  "vector-set!",
  "vector-fill!",
  /* 6.4. Control features */
  "procedure?",
  /* 6.5. Eval */
  "eval",
  /* 6.6. Input and output */
  /* 6.6.1. Ports */
  "input-port?",
  "output-port?",
  "current-input-port",
  "current-output-port",
  "open-input-file",
  "open-output-file",
  "close-input-port",
  "close-output-port",
  /* 6.6.2. Input */
  "read-char",
  "peek-char",
  "eof-object?",
  /* 6.6.3. Output */
  "write-char",
  /* 6.6.4. System interface */
  /* "load", */
  /* my extensions */
  "snarf-file"
};

struct scm_thing * interactive_variables;

struct scm_thing * create_library_variables(struct scm_thing * (**procedures)(struct scm_thing *),
					    const char ** procedure_names,
					    uint32_t procedure_count,
					    struct scm_thing * (**macros)(struct scm_thing *,
									  struct scm_thing *,
									  struct scm_thing *),
					    const char ** macro_names,
					    uint32_t macro_count,
					    struct scm_thing * variables) {
  for (uint32_t i = 0; i < procedure_count; i++) {
    struct scm_thing * procedure = malloc(sizeof(struct scm_thing));
    procedure->type = COMPILED_PROCEDURE;
    procedure->mutable = false;
    procedure->procedure = procedures[i];
    struct scm_thing * procedure_symbol = malloc(sizeof(struct scm_thing));
    procedure_symbol->type = SYMBOL;
    procedure_symbol->mutable = false;
    procedure_symbol->symbol = procedure_names[i];
    variables = c_cons(c_cons(procedure_symbol, procedure),
		       variables);
  }
  for (uint32_t i = 0; i < macro_count; i++) {
    struct scm_thing * macro = malloc(sizeof(struct scm_thing));
    macro->type = COMPILED_MACRO;
    macro->mutable = false;
    macro->macro = macros[i];
    struct scm_thing * macro_symbol = malloc(sizeof(struct scm_thing));
    macro_symbol->type = SYMBOL;
    macro_symbol->mutable = false;
    macro_symbol->symbol = macro_names[i];
    variables = c_cons(c_cons(macro_symbol, macro),
		       variables);
  }
  return variables;
}

void create_null_variables(void) {
  null_variables = create_library_variables(NULL,
					    NULL,
					    0,
					    null_macros,
					    null_macro_names,
					    sizeof(null_macros) / sizeof(null_macros[0]),
					    NULL);
  null_variables = c_cons(c_cons(&scm_undefined_symbol,
				 &scm_undefined),
			  null_variables);
}

void create_interactive_variables(void) {
  interactive_variables = create_library_variables(interactive_procedures,
						   interactive_procedure_names,
						   sizeof(interactive_procedures)
						   / sizeof(interactive_procedures[0]),
						   interactive_macros,
						   interactive_macro_names,
						   sizeof(interactive_macros)
						   / sizeof(interactive_macros[0]),
						   null_variables);
}
