#include "standard.h"

#include "eval.h"
#include "program.h"
#include "library.h"
#include "primitive.h"

#include <string.h>
#include <math.h>
#include <limits.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>

void require_exact_args(struct scm_thing * args, int32_t n) {
  if (c_length(args) != n) {
    printf("needed %i args but %i were given\n", n, c_length(args));
    error();
  }
}


/* 6.1. Equivalence predicates */

bool c_eqv_p(const struct scm_thing * x,
	     const struct scm_thing * y) {
  if (x == y) {
    return true;
  } else if (!x || !y) {
    return (!x && !y);
  } else if (x->type != y->type) {
#if 0
    if ((x->type == REAL_RATIONAL &&
	 y->type == COMPLEX_RATIONAL &&
	 y->imaginarynumerator == 0 &&
	 x->realnumerator == y->realnumerator &&
	 x->realdenominator == y->realdenominator) ||
	(x->type == COMPLEX_RATIONAL &&
	 y->type == REAL_RATIONAL &&
	 x->imaginarynumerator == 0 &&
	 x->realnumerator == y->realnumerator &&
	 x->realdenominator == y->realdenominator)) {
      // i mean really the complex one should be rational
      return true;
    }
#endif
    return false;
  } else {
    switch (x->type) {
    case NONE:
      return false;
    case PAIR:
      if (x->car == y->car &&
	  x->cdr == y->cdr) {
	return true;
      } else {
	return false;
      }
    case BOOLEAN:
      if (x->boolean == y->boolean) {
	return true;
      } else {
	return false;
      }
    case SYMBOL:
      if (!strcmp(x->symbol, y->symbol)) {
	return true;
      } else {
	return false;
      }
    case REAL_RATIONAL:
      if (x->realnumerator == y->realnumerator &&
	  (x->realnumerator == 0 ||
	   x->realdenominator == y->realdenominator)) {
	return true;
      } else {
	return false;
      }
    case REAL_FLOAT:
      if (x->realfloat == y->realfloat) {
	return true;
      } else {
	return false;
      }
    case COMPLEX_RATIONAL:
      if (x->realnumerator == y->realnumerator &&
	  (x->realnumerator == 0 ||
	   x->realdenominator == y->realdenominator) &&
	  x->imaginarynumerator == y->imaginarynumerator &&
	  (x->imaginarynumerator == 0 ||
	   x->imaginarydenominator == y->imaginarydenominator)) {
	return true;
      } else {
	return false;
      }
    case COMPLEX_FLOAT:
      if (x->realfloat == y->realfloat &&
	  x->imaginaryfloat == y->imaginaryfloat) {
	return true;
      } else {
	return false;
      }
    case CHARACTER:
      if (x->character == y->character) {
	return true;
      } else {
	return false;
      }
    case STRING:
      if (x->string == y->string) {
	return true;
      } else {
	return false;
      }
    case VECTOR:
      if (x->vectorsize == y->vectorsize &&
	  x->vector == y->vector) {
	return true;
      } else {
	return false;
      }
    case INPUT_PORT:
    case OUTPUT_PORT:
      // you shouldn't have 2 ports on the same fd
      return false;
    case COMPILED_PROCEDURE:
      if (x->procedure == y->procedure) {
	return true;
      } else {
	return false;
      }
    case INTERPRETED_PROCEDURE:
      if (x->car == y->car &&
	  x->cdr == y->cdr) {
	return true;
      } else {
	return false;
      }
    case COMPILED_MACRO:
      if (x->macro == y->macro) {
	return true;
      } else {
	return false;
      }
    case INTERPRETED_MACRO:
      if (x->car == y->car &&
	  x->cdr == y->cdr) {
	return true;
      } else {
	return false;
      }
    case INDIRECTION:
      // needed for the way we do closure
      if (x->car == y->car) {
	return true;
      } else {
	return false;
      }
    case UNDEFINED:
      error();
    }
  }
}

struct scm_thing * scm_eqv_p(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  if (c_eqv_p(args->car, args->cdr->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

bool c_eq_p(const struct scm_thing * x,
	    const struct scm_thing * y) {
  // this can just be the same as eqv?
  return c_eqv_p(x, y);
}

struct scm_thing * scm_eq_p(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  if (c_eq_p(args->car, args->cdr->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

bool c_equal_p(const struct scm_thing * x,
	       const struct scm_thing * y) {
  if (x == y) {
    return true;
  } else if (!x || !y) {
    return (!x && !y);
  } else if (x->type != y->type) {
    return false;
  } else {
    switch (x->type) {
    case PAIR:
      if (x->car == y->car &&
	  x->cdr == y->cdr) {
	return true;
      } else {
	if (c_equal_p(x->car, y->car)) {
	  return c_equal_p(x->cdr, y->cdr);
	} else {
	  return false;
	}
      }
    case STRING:
      if (x->stringsize == y->stringsize &&
	  (x->string == y->string ||
	   !memcmp(x->string, y->string, x->stringsize))) {
	return true;
      } else {
	return false;
      }
    case VECTOR:
      if (x->vectorsize != y->vectorsize) {
	return false;
      } else if (x->vector == y->vector) {
	return true;
      } else {
	for (uint32_t i = 0; i < x->vectorsize - 1; i++) {
	  if (!c_equal_p(x->vector[i], y->vector[i])) {
	    return false;
	  }
	}
	return c_equal_p(x->vector[x->vectorsize - 1], y->vector[x->vectorsize - 1]);
      }
    default:
      return c_eqv_p(x, y);
    }
  }
}

struct scm_thing * scm_equal_p(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  if (c_equal_p(args->car, args->cdr->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

/* 6.2.5. Numerical operations */

bool c_number_p(struct scm_thing * obj) {
  if (!obj) {
    return false;
  } else {
    switch (obj->type) {
    case REAL_RATIONAL:
    case REAL_FLOAT:
    case COMPLEX_RATIONAL:
    case COMPLEX_FLOAT:
      return true;
    default:
      return false;
    }
  }
}

struct scm_thing * scm_number_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (c_number_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

bool c_complex_p(struct scm_thing * obj) {
  return c_number_p(obj);
}

struct scm_thing * scm_complex_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (c_complex_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

bool c_real_p(struct scm_thing * obj) {
  if (!obj) {
    return false;
  } else {
    switch (obj->type) {
    case REAL_RATIONAL:
    case REAL_FLOAT:
      return true;
    case COMPLEX_RATIONAL:
      if (obj->imaginarynumerator == 0) {
	return true;
      } else {
	return false;
      }
    case COMPLEX_FLOAT:
      if (obj->imaginaryfloat == 0) {
	return true;
      } else {
	return false;
      }
    default:
      return false;
    }
  }
}

struct scm_thing * scm_real_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (c_real_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

bool c_rational_p(struct scm_thing * obj) {
  // floats are rational
  return c_real_p(obj);
}

struct scm_thing * scm_rational_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (c_rational_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

bool c_exact_integer_p(struct scm_thing * obj) {
  if (!obj) {
    return false;
  } else {
    switch (obj->type) {
    case REAL_RATIONAL:
      if (obj->realdenominator == 1) {
	return true;
      } else {
	return false;
      }
    default:
      return false;
    }
  }
}

bool c_integer_p(struct scm_thing * obj) {
  if (!obj) {
    return false;
  } else {
    switch (obj->type) {
    case REAL_RATIONAL:
      if (obj->realdenominator == 1) {
	return true;
      } else if ((obj->realdenominator % obj->realdenominator) == 0) {
	return true;
      } else {
	return false;
      }
    case REAL_FLOAT:
      if (floorf(obj->realfloat) == obj->realfloat) {
	return true;
      } else {
	return false;
      }
    case COMPLEX_RATIONAL:
      if (obj->imaginarynumerator == 0) {
	if (obj->realdenominator == 1) {
	  return true;
	} else if ((obj->realdenominator % obj->realdenominator) == 0) {
	  return true;
	} else {
	  return false;
	}
      } else {
	return false;
      }
    case COMPLEX_FLOAT:
      if (obj->imaginaryfloat == 0 &&
	  floorf(obj->realfloat) == obj->realfloat) {
	return true;
      } else {
	return false;
      }
    default:
      return false;
    }
  }
}

struct scm_thing * scm_integer_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (c_integer_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

bool c_exact_p(struct scm_thing * obj) {
  if (!obj) {
    return false;
  } else {
    switch (obj->type) {
    case REAL_RATIONAL:
    case COMPLEX_RATIONAL:
      return true;
    default:
      return false;
    }
  }
}

struct scm_thing * scm_exact_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (c_exact_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

bool c_inexact_p(struct scm_thing * obj) {
  if (!obj) {
    return false;
  } else {
    switch (obj->type) {
    case REAL_FLOAT:
    case COMPLEX_FLOAT:
      return true;
    default:
      return false;
    }
  }
}

struct scm_thing * scm_inexact_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (c_exact_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

// screw this, no more complex numbers!
bool c_numerical_equal(struct scm_thing * z1,
		       struct scm_thing * z2) {
  if (!z1 || !z2) {
    error();
  }
  switch (z1->type) {
  case REAL_RATIONAL:
    switch (z2->type) {
    case REAL_RATIONAL:
      // need promotions
      int64_t a = z1->realnumerator;
      uint64_t b = z1->realdenominator;
      int64_t c = z2->realnumerator;
      uint64_t d = z2->realdenominator;
      if (a * d == b * c) {
	return true;
      } else {
	return false;
      }
      break;
    case REAL_FLOAT:
      // no
      return false;
    default:
      error();
    }
  case REAL_FLOAT:
    switch (z2->type) {
    case REAL_RATIONAL:
      // no
      return false;
    case REAL_FLOAT:
      if (z1->realfloat == z2->realfloat) {
	return true;
      } else {
	return false;
      }
    default:
      error();
    }
  default:
    error();
  }
}

struct scm_thing * scm_numerical_equal(struct scm_thing * args) {
  if (c_length(args) < 2) {
    error();
  }
  while (args &&
	 args->type == PAIR &&
	 args->cdr &&
	 args->cdr->type == PAIR) {
    if (!c_numerical_equal(args->car, args->cdr->car)) {
      return &scm_false;
    } else {
      args = args->cdr;
    }
  }
  return &scm_true;
}

bool c_numerical_less_than(struct scm_thing * x1,
			   struct scm_thing * x2) {
  if (!x1 || !x2) {
    error();
  }
  switch (x1->type) {
  case REAL_RATIONAL:
    switch (x2->type) {
    case REAL_RATIONAL:
      // need promotions
      int64_t a = x1->realnumerator;
      uint64_t b = x1->realdenominator;
      int64_t c = x2->realnumerator;
      uint64_t d = x2->realdenominator;
      if (a * d < b * c) {
	return true;
      } else {
	return false;
      }
      break;
    case REAL_FLOAT:
      double ad = x1->realnumerator;
      double bd = x1->realdenominator;
      if (ad / bd < x2->realfloat) {
	return true;
      } else {
	return false;
      }
      break;
    default:
      error();
    }
  case REAL_FLOAT:
    switch (x2->type) {
    case REAL_RATIONAL:
      double cd = x2->realnumerator;
      double dd = x2->realdenominator;
      if (x1->realfloat < cd / dd) {
	return true;
      } else {
	return false;
      }
    case REAL_FLOAT:
      if (x1->realfloat < x2->realfloat) {
	return true;
      } else {
	return false;
      }
    default:
      error();
    }
  default:
    error();
  }
}

struct scm_thing * scm_numerical_less_than(struct scm_thing * args) {
  if (c_length(args) < 2) {
    error();
  }
  while (args &&
	 args->type == PAIR &&
	 args->cdr &&
	 args->cdr->type == PAIR) {
    if (!c_numerical_less_than(args->car, args->cdr->car)) {
      return &scm_false;
    } else {
      args = args->cdr;
    }
  }
  return &scm_true;
}

struct scm_thing * scm_numerical_greater_than_or_equal_to(struct scm_thing * args) {
  if (c_length(args) < 2) {
    error();
  }
  while (args &&
	 args->type == PAIR &&
	 args->cdr &&
	 args->cdr->type == PAIR) {
    if (c_numerical_less_than(args->car, args->cdr->car)) {
      return &scm_false;
    } else {
      args = args->cdr;
    }
  }
  return &scm_true;
}

bool c_numerical_greater_than(struct scm_thing * x1,
			      struct scm_thing * x2) {
  if (!x1 || !x2) {
    error();
  }
  switch (x1->type) {
  case REAL_RATIONAL:
    switch (x2->type) {
    case REAL_RATIONAL:
      // need promotions
      int64_t a = x1->realnumerator;
      uint64_t b = x1->realdenominator;
      int64_t c = x2->realnumerator;
      uint64_t d = x2->realdenominator;
      if (a * d > b * c) {
	return true;
      } else {
	return false;
      }
      break;
    case REAL_FLOAT:
      double ad = x1->realnumerator;
      double bd = x1->realdenominator;
      if (ad / bd > x2->realfloat) {
	return true;
      } else {
	return false;
      }
      break;
    default:
      error();
    }
  case REAL_FLOAT:
    switch (x2->type) {
    case REAL_RATIONAL:
      double cd = x2->realnumerator;
      double dd = x2->realdenominator;
      if (x1->realfloat > cd / dd) {
	return true;
      } else {
	return false;
      }
      break;
    case REAL_FLOAT:
      if (x1->realfloat > x2->realfloat) {
	return true;
      } else {
	return false;
      }
    default:
      error();
    }
  default:
    error();
  }
}

struct scm_thing * scm_numerical_greater_than(struct scm_thing * args) {
  if (c_length(args) < 2) {
    error();
  }
  while (args &&
	 args->type == PAIR &&
	 args->cdr &&
	 args->cdr->type == PAIR) {
    if (!c_numerical_greater_than(args->car, args->cdr->car)) {
      return &scm_false;
    } else {
      args = args->cdr;
    }
  }
  return &scm_true;
}

struct scm_thing * scm_numerical_less_than_or_equal_to(struct scm_thing * args) {
  if (c_length(args) < 2) {
    error();
  }
  while (args &&
	 args->type == PAIR &&
	 args->cdr &&
	 args->cdr->type == PAIR) {
    if (c_numerical_greater_than(args->car, args->cdr->car)) {
      return &scm_false;
    } else {
      args = args->cdr;
    }
  }
  return &scm_true;
}

struct scm_thing * scm_zero_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * z = args->car;
  if (!z) {
    error();
  }
  switch (z->type) {
  case REAL_RATIONAL:
    if (z->realnumerator == 0) {
      return &scm_true;
    } else {
      return &scm_false;
    }
  case REAL_FLOAT:
    if (z->realfloat == 0.0f) {
      return &scm_true;
    } else {
      return &scm_false;
    }
  default:
    error();
  }
}

struct scm_thing * scm_positive_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * x = args->car;
  if (!x) {
    error();
  }
  switch (x->type) {
  case REAL_RATIONAL:
    if (x->realnumerator > 0) {
      return &scm_true;
    } else {
      return &scm_false;
    }
  case REAL_FLOAT:
    if (x->realfloat > 0.0f) {
      return &scm_true;
    } else {
      return &scm_false;
    }
  default:
    error();
  }
}

struct scm_thing * scm_negative_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * x = args->car;
  if (!x) {
    error();
  }
  switch (x->type) {
  case REAL_RATIONAL:
    if (x->realnumerator < 0) {
      return &scm_true;
    } else {
      return &scm_false;
    }
  case REAL_FLOAT:
    if (x->realfloat < 0.0f) {
      return &scm_true;
    } else {
      return &scm_false;
    }
  default:
    error();
  }
}

struct scm_thing * scm_odd_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * n = args->car;
  if (!n) {
    error();
  }
  switch (n->type) {
  case REAL_RATIONAL:
    if (n->realdenominator == 1 &&
	n->realnumerator % 2) {
      return &scm_true;
    } else {
      return &scm_false;
    }
  default:
    error();
  }
}

struct scm_thing * scm_even_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * n = args->car;
  if (!n) {
    error();
  }
  switch (n->type) {
  case REAL_RATIONAL:
    if (n->realdenominator == 1 &&
	!(n->realnumerator % 2)) {
      return &scm_true;
    } else {
      return &scm_false;
    }
  default:
    error();
  }
}

uint64_t c_gcd(uint64_t n,
	       uint64_t m) {
  if (n % m) {
    return c_gcd(m, n % m);
  } else {
    return m;
  }
}

struct scm_thing * create_rational(int64_t n,
				   uint64_t d) {
  uint64_t commonfactor;
  if (n == 0) {
    return &scm_zero;
  } else if (n > 0) {
    commonfactor = c_gcd(n, d);
  } else {
    commonfactor = c_gcd(-n, d);
  }
  n = n / commonfactor;
  d = d / commonfactor;
  if (n > INT32_MAX || n < INT32_MIN) {
    // violation of implementation restriction
    error();
  } else if (d > UINT32_MAX) {
    // violation of implementation restriction
    error();
  } else {
    struct scm_thing * rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_RATIONAL;
    rv->mutable = true;
    rv->realnumerator = n;
    rv->realdenominator = d;
    return rv;
  }
}

struct scm_thing * create_integer(int32_t n) {
  struct scm_thing * rv = malloc(sizeof(struct scm_thing));
  rv->type = REAL_RATIONAL;
  rv->mutable = true;
  rv->realnumerator = n;
  rv->realdenominator = 1;
  return rv;
}

struct scm_thing * c_add(struct scm_thing * z1,
			 struct scm_thing * z2) {
  if (!z1 || !z2) {
    error();
  }
  struct scm_thing * rv;
  switch (z1->type) {
  case REAL_RATIONAL:
    switch (z2->type) {
    case REAL_RATIONAL:
      // need promotions
      int64_t a = z1->realnumerator;
      uint64_t b = z1->realdenominator;
      int64_t c = z2->realnumerator;
      uint64_t d = z2->realdenominator;
      return create_rational(a * d + c * d,
			     b * d);
    case REAL_FLOAT:
      double ad = z1->realnumerator;
      double bd = z1->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = (ad / bd) + z2->realfloat;
      return rv;
    default:
      error();
    }
  case REAL_FLOAT:
    switch (z2->type) {
    case REAL_RATIONAL:
      double cd = z2->realnumerator;
      double dd = z2->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = z1->realfloat + (cd / dd);
      return rv;
    case REAL_FLOAT:
      struct scm_thing * rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = z1->realfloat + z2->realfloat;
      return rv;
    default:
      error();
    }
  default:
    error();
  }
}

struct scm_thing * scm_add(struct scm_thing * args) {
  if (c_length(args) < 0) {
    error();
  } else if (c_length(args) == 0) {
    return &scm_zero;
  } else if (c_length(args) == 1) {
    return args->car;
  } else {
    struct scm_thing * sum;
    sum = c_add(args->car, args->cdr->car);
    args = args->cdr->cdr;
    while (args) {
      struct scm_thing * newsum;
      newsum = c_add(sum, args->car);
      // for once we don't leak memory everywhere
      if (sum->mutable) {
	free(sum);
      }
      sum = newsum;
      args = args->cdr;
    }
    return sum;
  }
}

struct scm_thing * c_subtract(struct scm_thing * z1,
			      struct scm_thing * z2) {
  if (!z1 || !z2) {
    error();
  }
  struct scm_thing * rv;
  switch (z1->type) {
  case REAL_RATIONAL:
    switch (z2->type) {
    case REAL_RATIONAL:
      // need promotions
      int64_t a = z1->realnumerator;
      uint64_t b = z1->realdenominator;
      int64_t c = z2->realnumerator;
      uint64_t d = z2->realdenominator;
      return create_rational(a * d - c * d,
			     b * d);
    case REAL_FLOAT:
      double ad = z1->realnumerator;
      double bd = z1->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = (ad / bd) - z2->realfloat;
      return rv;
    default:
      error();
    }
  case REAL_FLOAT:
    switch (z2->type) {
    case REAL_RATIONAL:
      double cd = z2->realnumerator;
      double dd = z2->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = z1->realfloat - (cd / dd);
      return rv;
    case REAL_FLOAT:
      struct scm_thing * rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = z1->realfloat - z2->realfloat;
      return rv;
    default:
      error();
    }
  default:
    error();
  }
}

struct scm_thing * scm_subtract(struct scm_thing * args) {
  if (c_length(args) < 1) {
    error();
  } else if (c_length(args) == 1) {
    struct scm_thing * z = args->car;
    if (!z) {
      error();
    }
    struct scm_thing * rv;
    switch (z->type) {
    case REAL_RATIONAL:
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_RATIONAL;
      rv->mutable = true;
      rv->realnumerator = -z->realnumerator;
      rv->realdenominator = z->realdenominator;
      return rv;
    case REAL_FLOAT:
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = -z->realfloat;
      return rv;
    default:
      error();
    }
  } else {
    struct scm_thing * difference;
    difference = c_subtract(args->car, args->cdr->car);
    args = args->cdr->cdr;
    while (args) {
      struct scm_thing * newdifference;
      newdifference = c_subtract(difference, args->car);
      // for once we don't leak memory everywhere
      if (difference->mutable) {
	free(difference);
      }
      difference = newdifference;
      args = args->cdr;
    }
    return difference;
  }
}

struct scm_thing * c_multiply(struct scm_thing * z1,
			      struct scm_thing * z2) {
  if (!z1 || !z2) {
    error();
  }
  if (z1 == &scm_zero ||
      z2 == &scm_zero) {
    // anything times zero is zero
    // maybe even things that aren't numbers?
    return &scm_zero;
  }
  struct scm_thing * rv;
  switch (z1->type) {
  case REAL_RATIONAL:
    switch (z2->type) {
    case REAL_RATIONAL:
      // need promotions
      int64_t a = z1->realnumerator;
      uint64_t b = z1->realdenominator;
      int64_t c = z2->realnumerator;
      uint64_t d = z2->realdenominator;
      return create_rational(a * c,
			     b * d);
    case REAL_FLOAT:
      double ad = z1->realnumerator;
      double bd = z1->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = (ad / bd) * z2->realfloat;
      return rv;
    default:
      error();
    }
  case REAL_FLOAT:
    switch (z2->type) {
    case REAL_RATIONAL:
      double cd = z2->realnumerator;
      double dd = z2->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = z1->realfloat * (cd / dd);
      return rv;
    case REAL_FLOAT:
      struct scm_thing * rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = z1->realfloat * z2->realfloat;
      return rv;
    default:
      error();
    }
  default:
    error();
  }
}

struct scm_thing * scm_multiply(struct scm_thing * args) {
  if (c_length(args) < 0) {
    error();
  } else if (c_length(args) == 0) {
    return &scm_one;
  } else if (c_length(args) == 1) {
    return args->car;
  } else {
    struct scm_thing * product;
    product = c_multiply(args->car, args->cdr->car);
    args = args->cdr->cdr;
    while (args) {
      if (product == &scm_zero) {
	// prevents attempting to free a constant
	return product;
      }
      struct scm_thing * newproduct;
      newproduct = c_multiply(product, args->car);
      // for once we don't leak memory everywhere
      if (product != &scm_one) {
	free(product);
      }
      product = newproduct;
      args = args->cdr;
    }
    return product;
  }
}

struct scm_thing * c_divide(struct scm_thing * z1,
			    struct scm_thing * z2) {
  if (!z1 || !z2) {
    error();
  }
  if (z2 == &scm_zero) {
    // not allowed
    error();
  }
  if (z1 == &scm_zero) {
    // zero divided by anything nonzero is zero
    return &scm_zero;
  }
  struct scm_thing * rv;
  switch (z1->type) {
  case REAL_RATIONAL:
    switch (z2->type) {
    case REAL_RATIONAL:
      // need promotions
      int64_t a = z1->realnumerator;
      uint64_t b = z1->realdenominator;
      int64_t c = z2->realnumerator;
      uint64_t d = z2->realdenominator;
      if (c == 0) {
	error();
      } else if (c > 0) {
	return create_rational(a * d,
			       b * c);
      } else {
	return create_rational(-a * d,
			       b * -c);
      }
    case REAL_FLOAT:
      double ad = z1->realnumerator;
      double bd = z1->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = (ad / bd) / z2->realfloat;
      return rv;
    default:
      error();
    }
  case REAL_FLOAT:
    switch (z2->type) {
    case REAL_RATIONAL:
      double cd = z2->realnumerator;
      double dd = z2->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = z1->realfloat / (cd / dd);
      return rv;
    case REAL_FLOAT:
      struct scm_thing * rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = z1->realfloat / z2->realfloat;
      return rv;
    default:
      error();
    }
  default:
    error();
  }
}

struct scm_thing * scm_divide(struct scm_thing * args) {
  if (c_length(args) < 1) {
    error();
  } else if (c_length(args) == 1) {
    struct scm_thing * z = args->car;
    if (!z) {
      error();
    }
    struct scm_thing * rv;
    switch (z->type) {
    case REAL_RATIONAL:
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_RATIONAL;
      rv->mutable = true;
      if (z->realnumerator == 0) {
	error();
      } else if (z->realnumerator > 0) {
	rv->realnumerator = z->realdenominator;
	rv->realdenominator = z->realnumerator;
      } else {
	rv->realnumerator = -z->realdenominator;
	rv->realdenominator = -z->realnumerator;
      }
      return rv;
    case REAL_FLOAT:
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = 1 / z->realfloat;
      return rv;
    default:
      error();
    }
  } else {
    struct scm_thing * quotient;
    quotient = c_divide(args->car, args->cdr->car);
    args = args->cdr->cdr;
    while (args) {
      if (quotient == &scm_zero) {
	// prevents attempting to free a constant
	return quotient;
      }
      struct scm_thing * newquotient;
      newquotient = c_divide(quotient, args->car);
      // for once we don't leak memory everywhere
      if (quotient != &scm_one) {
	free(quotient);
      }
      quotient = newquotient;
      args = args->cdr;
    }
    return quotient;
  }
}

struct scm_thing * scm_abs(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * x = args->car;
  if (!x) {
    error();
  }
  struct scm_thing * rv;
  switch (x->type) {
  case REAL_RATIONAL:
    if (x->realnumerator >= 0) {
      return x;
    } else {
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_RATIONAL;
      rv->mutable = true;
      rv->realnumerator = -x->realnumerator;
      rv->realdenominator = x->realdenominator;
      return rv;
    }
  case REAL_FLOAT:
    if (x->realfloat >= 0.0f) {
      return x;
    } else {
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = -x->realfloat;
      return rv;
    }
  default:
    error();
  }
}

struct scm_thing * scm_quotient(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * n1 = args->car;
  struct scm_thing * n2 = args->cdr->car;
  if (!n1 || !n2) {
    error();
  }
  if (n1->type != REAL_RATIONAL ||
      n2->type != REAL_RATIONAL) {
    error();
  }
  if (n1->realdenominator != 1 ||
      n2->realdenominator != 1) {
    error();
  }
  if (n2->realnumerator == 0) {
    error();
  }
  struct scm_thing * rv = malloc(sizeof(struct scm_thing));
  rv->type = REAL_RATIONAL;
  rv->mutable = true;
  rv->realdenominator = 1;
  if (n1->realnumerator >= 0) {
    if (n2->realnumerator >= 0) {
      rv->realnumerator = n1->realnumerator / n2->realnumerator;
    } else {
      rv->realnumerator = -(n1->realnumerator / (-n2->realnumerator));
    }
  } else {
    if (n2->realnumerator >= 0) {
      rv->realnumerator = -((-n1->realnumerator) / n2->realnumerator);
    } else {
      rv->realnumerator = (-n1->realnumerator) / (-n2->realnumerator);
    }
  }
  return rv;
}

struct scm_thing * scm_remainder(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * n1 = args->car;
  struct scm_thing * n2 = args->cdr->car;
  if (!n1 || !n2) {
    error();
  }
  if (n1->type != REAL_RATIONAL ||
      n2->type != REAL_RATIONAL) {
    error();
  }
  if (n1->realdenominator != 1 ||
      n2->realdenominator != 1) {
    error();
  }
  if (n2->realnumerator == 0) {
    error();
  }
  struct scm_thing * rv = malloc(sizeof(struct scm_thing));
  rv->type = REAL_RATIONAL;
  rv->mutable = true;
  rv->realdenominator = 1;
  // same sign as n1
  if (n1->realnumerator >= 0) {
    if (n2->realnumerator >= 0) {
      rv->realnumerator = n1->realnumerator % n2->realnumerator;
    } else {
      rv->realnumerator = n1->realnumerator % (-n2->realnumerator);
    }
  } else {
    if (n2->realnumerator >= 0) {
      rv->realnumerator = -((-n1->realnumerator) % n2->realnumerator);
    } else {
      rv->realnumerator = -((-n1->realnumerator) % (-n2->realnumerator));
    }
  }
  return rv;
}

struct scm_thing * scm_modulo(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * n1 = args->car;
  struct scm_thing * n2 = args->cdr->car;
  if (!n1 || !n2) {
    error();
  }
  if (n1->type != REAL_RATIONAL ||
      n2->type != REAL_RATIONAL) {
    error();
  }
  if (n1->realdenominator != 1 ||
      n2->realdenominator != 1) {
    error();
  }
  if (n2->realnumerator == 0) {
    error();
  }
  struct scm_thing * rv = malloc(sizeof(struct scm_thing));
  rv->type = REAL_RATIONAL;
  rv->mutable = true;
  rv->realdenominator = 1;
  // same sign as n2
  if (n1->realnumerator >= 0) {
    if (n2->realnumerator >= 0) {
      rv->realnumerator = n1->realnumerator % n2->realnumerator;
    } else {
      rv->realnumerator = (n1->realnumerator % (-n2->realnumerator)) + n2->realnumerator;
    }
  } else {
    if (n2->realnumerator >= 0) {
      rv->realnumerator = -((-n1->realnumerator) % n2->realnumerator) + n2->realnumerator;
    } else {
      rv->realnumerator = -((-n1->realnumerator) % (-n2->realnumerator));
    }
  }
  return rv;
}

struct scm_thing * scm_numerator(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * q = args->car;
  if (!q) {
    error();
  }
  switch (q->type) {
  case REAL_RATIONAL:
    struct scm_thing * rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_RATIONAL;
    rv->mutable = true;
    rv->realnumerator = q->realnumerator;
    rv->realdenominator = 1;
    return rv;
  default:
    // screw you you aren't getting the numerator of a float
    error();
  }
}

struct scm_thing * scm_denominator(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * q = args->car;
  if (!q) {
    error();
  }
  switch (q->type) {
  case REAL_RATIONAL:
    struct scm_thing * rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_RATIONAL;
    rv->mutable = true;
    rv->realnumerator = q->realdenominator;
    rv->realdenominator = 1;
    return rv;
  default:
    // screw you you aren't getting the numerator of a float
    error();
  }
}

struct scm_thing * scm_floor(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * x = args->car;
  if (!x) {
    error();
  }
  struct scm_thing * rv;
  switch (x->type) {
  case REAL_RATIONAL:
    if (x->realdenominator == 1) {
      return x;
    } else {
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_RATIONAL;
      rv->mutable = true;
      if (x->realnumerator >= 0) {
	rv->realnumerator = x->realnumerator / x->realdenominator;
      } else {
	// c division rounds towards zero
	rv->realnumerator = -((-x->realnumerator) / x->realdenominator);
      }
      rv->realdenominator = 1;
      return rv;
    }
  case REAL_FLOAT:
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    rv->realfloat = floor(x->realfloat);
    return rv;
  default:
    error();
  }
}

struct scm_thing * scm_ceiling(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * x = args->car;
  if (!x) {
    error();
  }
  struct scm_thing * rv;
  switch (x->type) {
  case REAL_RATIONAL:
    if (x->realdenominator == 1) {
      return x;
    } else {
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_RATIONAL;
      rv->mutable = true;
      if (x->realnumerator >= 0) {
	// c division rounds towards zero
	rv->realnumerator = -((-x->realnumerator) / x->realdenominator);
      } else {
	rv->realnumerator = x->realnumerator / x->realdenominator;
      }
      rv->realdenominator = 1;
      return rv;
    }
  case REAL_FLOAT:
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    rv->realfloat = ceil(x->realfloat);
    return rv;
  default:
    error();
  }
}

struct scm_thing * scm_truncate(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * x = args->car;
  if (!x) {
    error();
  }
  struct scm_thing * rv;
  switch (x->type) {
  case REAL_RATIONAL:
    if (x->realdenominator == 1) {
      return x;
    } else {
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_RATIONAL;
      rv->mutable = true;
      // c division rounds towards zero
      rv->realnumerator = x->realnumerator / x->realdenominator;
      rv->realdenominator = 1;
      return rv;
    }
  case REAL_FLOAT:
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    rv->realfloat = trunc(x->realfloat);
    return rv;
  default:
    error();
  }
}

struct scm_thing * scm_round(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * x = args->car;
  if (!x) {
    error();
  }
  struct scm_thing * rv;
  switch (x->type) {
  case REAL_RATIONAL:
    if (x->realdenominator == 1) {
      return x;
    } else {
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_RATIONAL;
      rv->mutable = true;
      uint32_t n;
      int32_t sign;
      if (x->realnumerator >= 0) {
	n = x->realnumerator;
	sign = 1;
      } else {
	n = -x->realnumerator;
	sign = -1;
      }
      if (x->realdenominator == 2) {
	// round towards even
	if (n % 4 == 1) {
	  rv->realnumerator = n / x->realdenominator;
	} else {
	  // x->realnumerator % 4 == 3
	  rv->realnumerator = 1 + (n / x->realdenominator);
	}
      } else {
	// will lie on one side or another
	if ((n % x->realdenominator) > (x->realdenominator / 2)) {
	  // round up
	  rv->realnumerator = 1 + (n / x->realdenominator);
	} else {
	  // round down
	  rv->realnumerator = n / x->realdenominator;
	}
      }
      rv->realnumerator = sign * rv->realnumerator;
      rv->realdenominator = 1;
      return rv;
    }
  case REAL_FLOAT:
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    // rounding mode should be FE_TONEAREST for even rounding
    rv->realfloat = nearbyint(x->realfloat);
    return rv;
  default:
    error();
  }
}

struct scm_thing * scm_exp(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * z = args->car;
  if (!z) {
    error();
  }
  struct scm_thing * rv;
  switch (z->type) {
  case REAL_RATIONAL:
    double ad = z->realnumerator;
    double bd = z->realdenominator;
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    rv->realfloat = exp(ad/bd);
    return rv;
  case REAL_FLOAT:
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    rv->realfloat = exp(z->realfloat);
    return rv;
  default:
    error();
  }
}

struct scm_thing * scm_log(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * z = args->car;
  if (!z) {
    error();
  }
  struct scm_thing * rv;
  switch (z->type) {
  case REAL_RATIONAL:
    if (z == &scm_one) {
      return &scm_zero;
    } else {
      double ad = z->realnumerator;
      double bd = z->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = log(ad/bd);
      return rv;
    }
  case REAL_FLOAT:
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    rv->realfloat = log(z->realfloat);
    return rv;
  default:
    error();
  }
}

struct scm_thing * scm_sin(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * z = args->car;
  if (!z) {
    error();
  }
  struct scm_thing * rv;
  switch (z->type) {
  case REAL_RATIONAL:
    if (z == &scm_zero) {
      return &scm_zero;
    } else {
      double ad = z->realnumerator;
      double bd = z->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = sin(ad/bd);
      return rv;
    }
  case REAL_FLOAT:
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    rv->realfloat = sin(z->realfloat);
    return rv;
  default:
    error();
  }
}

struct scm_thing * scm_cos(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * z = args->car;
  if (!z) {
    error();
  }
  struct scm_thing * rv;
  switch (z->type) {
  case REAL_RATIONAL:
    if (z == &scm_zero) {
      return &scm_one;
    } else {
      double ad = z->realnumerator;
      double bd = z->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = cos(ad/bd);
      return rv;
    }
  case REAL_FLOAT:
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    rv->realfloat = cos(z->realfloat);
    return rv;
  default:
    error();
  }
}

struct scm_thing * scm_tan(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * z = args->car;
  if (!z) {
    error();
  }
  struct scm_thing * rv;
  switch (z->type) {
  case REAL_RATIONAL:
    if (z == &scm_zero) {
      return &scm_zero;
    } else {
      double ad = z->realnumerator;
      double bd = z->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = tan(ad/bd);
      return rv;
    }
  case REAL_FLOAT:
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    rv->realfloat = tan(z->realfloat);
    return rv;
  default:
    error();
  }
}

struct scm_thing * scm_asin(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * z = args->car;
  if (!z) {
    error();
  }
  struct scm_thing * rv;
  switch (z->type) {
  case REAL_RATIONAL:
    if (z == &scm_zero) {
      return &scm_zero;
    } else {
      double ad = z->realnumerator;
      double bd = z->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = asin(ad/bd);
      return rv;
    }
  case REAL_FLOAT:
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    rv->realfloat = asin(z->realfloat);
    return rv;
  default:
    error();
  }
}

struct scm_thing * scm_acos(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * z = args->car;
  if (!z) {
    error();
  }
  struct scm_thing * rv;
  switch (z->type) {
  case REAL_RATIONAL:
    if (z == &scm_one) {
      return &scm_zero;
    } else {
      double ad = z->realnumerator;
      double bd = z->realdenominator;
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = acos(ad/bd);
      return rv;
    }
  case REAL_FLOAT:
    rv = malloc(sizeof(struct scm_thing));
    rv->type = REAL_FLOAT;
    rv->mutable = true;
    rv->realfloat = acos(z->realfloat);
    return rv;
  default:
    error();
  }
}

struct scm_thing * scm_atan(struct scm_thing * args) {
  if (c_length(args) < 1) {
    error();
  } else if (c_length(args) == 1) {
    struct scm_thing * z = args->car;
    if (!z) {
      error();
    }
    struct scm_thing * rv;
    switch (z->type) {
    case REAL_RATIONAL:
      if (z == &scm_zero) {
	return &scm_zero;
      } else {
	double ad = z->realnumerator;
	double bd = z->realdenominator;
	rv = malloc(sizeof(struct scm_thing));
	rv->type = REAL_FLOAT;
	rv->mutable = true;
	rv->realfloat = atan(ad/bd);
	return rv;
      }
    case REAL_FLOAT:
      rv = malloc(sizeof(struct scm_thing));
      rv->type = REAL_FLOAT;
      rv->mutable = true;
      rv->realfloat = atan(z->realfloat);
      return rv;
    default:
      error();
    }
  } else if (c_length(args) == 2) {
    struct scm_thing * y = args->car;
    struct scm_thing * x = args->cdr->car;
    if (!y || !x) {
      error();
    }
    struct scm_thing * rv;
    double ad, bd, cd, dd;
    switch (y->type) {
    case REAL_RATIONAL:
      ad = y->realnumerator;
      bd = y->realdenominator;
      switch (x->type) {
      case REAL_RATIONAL:
	// can't be bothered with exact values
	cd = x->realnumerator;
	dd = x->realdenominator;
	rv = malloc(sizeof(struct scm_thing));
	rv->type = REAL_FLOAT;
	rv->mutable = true;
	rv->realfloat = atan2(ad/bd, cd/dd);
	return rv;
      case REAL_FLOAT:
	rv = malloc(sizeof(struct scm_thing));
	rv->type = REAL_FLOAT;
	rv->mutable = true;
	rv->realfloat = atan2(ad/bd, x->realfloat);
	return rv;
      default:
	error();
      }
    case REAL_FLOAT:
      switch (x->type) {
      case REAL_RATIONAL:
	// can't be bothered with exact values
	cd = x->realnumerator;
	dd = x->realdenominator;
	rv = malloc(sizeof(struct scm_thing));
	rv->type = REAL_FLOAT;
	rv->mutable = true;
	rv->realfloat = atan2(y->realfloat, cd/dd);
	return rv;
      case REAL_FLOAT:
	rv = malloc(sizeof(struct scm_thing));
	rv->type = REAL_FLOAT;
	rv->mutable = true;
	rv->realfloat = atan2(y->realfloat, x->realfloat);
	return rv;
      default:
	error();
      }
    default:
      error();
    }
  } else {
    error();
  }
}

// can't really be bothered with the rest of the functions

/* 6.2.6. Numerical input and output */

// just realised that these implementations will not be very good,
// and I will have to write ones in scheme anyways

#if 0
char digit_to_char(uint32_t digit) {
  if (digit > 9) {
    return 0x61 + (digit - 10);
  } else {
    return 0x30 + digit;
  }
}

struct scm_thing * c_number_to_string(struct scm_thing * z,
				      uint32_t radix) {
  if (!z) {
    error();
  }
  uint32_t stringlength = 0;
  char buffer[65];
  switch (z->type) {
  case REAL_RATIONAL:
    uint32_t n;
    if (z->realnumerator < 0) {
      buffer[stringlength] = '-';
      stringlength++;
      n = -z->realnumerator;
    } else {
      n = z->realnumerator;
    }
    // very stupid way of doing this
    int32_t i = 0;
    uint32_t placevalue = 1;
    while (n / placevalue > radix) {
      i++;
      placevalue *= radix;
    }
    for (; i >= 0; i--) {
      buffer[stringlength] = digit_to_char(n / placevalue);
      stringlength++;
      n = n % placevalue;
      placevalue /= radix;
    }
    buffer[stringlength] = '/';
    stringlength++;
    // now again for the denominator
    n = z->realdenominator;
    i = 0;
    placevalue = 1;
    while (n / placevalue > radix) {
      i++;
      placevalue *= radix;
    }
    for (; i >= 0; i--) {
      buffer[stringlength] = digit_to_char(n / placevalue);
      stringlength++;
      n = n % placevalue;
      placevalue /= radix;
    }
    break;
  case REAL_FLOAT:
    // UNFINISHED
  }
}
#endif

/* 6.3. Other data types */

/* 6.3.1. Booleans */

struct scm_thing * scm_not(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * obj = args->car;
  if (obj && obj->type == BOOLEAN && obj->boolean == false) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_boolean_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * obj = args->car;
  if (obj && obj->type == BOOLEAN) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

/* 6.3.2. Pairs and lists */

struct scm_thing * scm_pair_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * obj = args->car;
  if (obj && obj->type == PAIR) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * c_cons(struct scm_thing *x,
			  struct scm_thing *y) {
  struct scm_thing * rv = malloc(sizeof(struct scm_thing));
  rv->type = PAIR;
  rv->mutable = true;
  rv->car = x;
  rv->cdr = y;
  return rv;
}

struct scm_thing * scm_cons(struct scm_thing * args) {
  require_exact_args(args, 2);
  return c_cons(args->car, args->cdr->car);
}

struct scm_thing * scm_car(struct scm_thing * args) {
  require_exact_args(args, 1);
  if (!args->car || args->car->type != PAIR) {
    printf("tried to take car of non-pair object\n");
    error();
  } else {
    return args->car->car;
  }
}

struct scm_thing * scm_cdr(struct scm_thing * args) {
  require_exact_args(args, 1);
  if (!args->car || args->car->type != PAIR) {
    printf("tried to take cdr of non-pair object\n");
    error();
  } else {
    return args->car->cdr;
  }
}

struct scm_thing * scm_set_car(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * pair = args->car;
  struct scm_thing * obj = args->cdr->car;
  if (!pair || pair->type != PAIR || pair->mutable == false) {
    error();
  } else {
    return pair->car = obj;
  }
}

struct scm_thing * scm_set_cdr(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * pair = args->car;
  struct scm_thing * obj = args->cdr->car;
  if (!pair || pair->type != PAIR || pair->mutable == false) {
    error();
  } else {
    return pair->cdr = obj;
  }
}

// can't be bothered with caar, cadr etc.

struct scm_thing * scm_null_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (!args->car) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_list_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * obj = args->car;
  if (!obj || c_length(obj) > 0) {
    // this works as if c_length is passed an object that is not a pair it returns 0
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_list(struct scm_thing * args) {
  // simplest procedure ever
  // note that this may not work correctly if you use apply rather than eval
  // TODO: check that
  return args;
}

int32_t c_length(struct scm_thing * list) {
  // a list large enough to cause an overflow would take up something like 2^34 bytes, or 16 GiB
  int rv = 0;
  while (list && list->type == PAIR) {
    rv++;
    list = list->cdr;
  }
  if (list) {
    // last element was not pair
    return -rv;
  }
  return rv;
}

struct scm_thing * scm_length(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  // if you use it on something that isn't a list,
  // it's your fault that it returns something unexpected:
  // garbage in, garbage out
  return create_integer(c_length(args->car));
}

// not sure I can be bothered with append and reverse

struct scm_thing * scm_append(struct scm_thing * args) {
  if (c_length(args) < 1) {
    error();
  }
  struct scm_thing * list_start = NULL;
  struct scm_thing * list_end = NULL;
  while (args->cdr) {
    struct scm_thing * old_list = args->car;
    while (old_list && old_list->type == PAIR) {
      if (list_end) {
	list_end->cdr = malloc(sizeof(struct scm_thing));
	list_end = list_end->cdr;
      } else {
	list_start = malloc(sizeof(struct scm_thing));
	list_end = list_start;
      }
      list_end->type = PAIR;
      list_end->mutable = true;
      list_end->car = old_list->car;
      old_list = old_list->cdr;
    }
    args = args->cdr;
  }
  if (list_end) {
    list_end->cdr = args->car;
    return list_start;
  } else {
    return args->car;
  }
}
    

// not sure why i need a c version but i'm making it anyways
// this one WILL error if you do something stupid
struct scm_thing * c_list_tail(struct scm_thing * list,
			       uint32_t k) {
  struct scm_thing * tail = list;
  for (uint32_t i = 0; i < k; i++) {
    if (!tail || tail->type != PAIR) {
      // doofus
      error();
    }
    tail = tail->cdr;
  }
  return tail;
}

struct scm_thing * scm_list_tail(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * list = args->car;
  struct scm_thing * k = args->cdr->car;
  if (k->type != REAL_RATIONAL) {
    // screw you if you try to use an inexact number
    error();
  }
  if (k->realdenominator != 1) {
    // what?
    // did you try to do integer division with "/"?
    error();
  }
  if (k->realnumerator < 0) {
    // this is not python
    error();
  }
  return c_list_tail(list, k->realnumerator);
}

struct scm_thing * c_memq(struct scm_thing * obj,
			  struct scm_thing * list) {
  while (list &&
	 list->type == PAIR) {
    if (c_eq_p(obj, list->car)) {
      return list;
    } else {
      list = list->cdr;
    }
  }
  return NULL;
}

struct scm_thing * c_memv(struct scm_thing * obj,
			  struct scm_thing * list) {
  while (list &&
	 list->type == PAIR) {
    if (c_eqv_p(obj, list->car)) {
      return list;
    } else {
      list = list->cdr;
    }
  }
  return NULL;
}

struct scm_thing * c_member(struct scm_thing * obj,
			    struct scm_thing * list) {
  while (list &&
	 list->type == PAIR) {
    if (c_equal_p(obj, list->car)) {
      return list;
    } else {
      list = list->cdr;
    }
  }
  return NULL;
}

struct scm_thing * scm_memq(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * rv;
  rv = c_memq(args->car, args->cdr->car);
  if (rv) {
    return rv;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_memv(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * rv;
  rv = c_memv(args->car, args->cdr->car);
  if (rv) {
    return rv;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_member(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * rv;
  rv = c_member(args->car, args->cdr->car);
  if (rv) {
    return rv;
  } else {
    return &scm_false;
  }
}

struct scm_thing * c_assoc(struct scm_thing * obj,
			   struct scm_thing * alist) {
  while (alist) {
    if (c_equal_p(obj, alist->car->car)) {
      return alist->car;
    }
    alist = alist->cdr;
  }
  return NULL;
}

// ok i'll write ones that give errors
struct scm_thing * scm_assq(struct scm_thing * args) {
  struct scm_thing * obj = args->car;
  struct scm_thing * alist = args->cdr->car;
  while (alist &&
	 alist->type == PAIR &&
	 alist->car->type == PAIR) {
    if (c_eq_p(obj, alist->car->car)) {
      return alist->car;
    }
    alist = alist->cdr;
  }
  if (alist) {
    printf("assq run on non-alist\n");
    error();
  }
  return &scm_false;
}

struct scm_thing * scm_assv(struct scm_thing * args) {
  struct scm_thing * obj = args->car;
  struct scm_thing * alist = args->cdr->car;
  while (alist &&
	 alist->type == PAIR &&
	 alist->car->type == PAIR) {
    if (c_eqv_p(obj, alist->car->car)) {
      return alist->car;
    }
    alist = alist->cdr;
  }
  if (alist) {
    printf("assv run on non-alist\n");
    error();
  }
  return &scm_false;
}

struct scm_thing * scm_assoc(struct scm_thing * args) {
  struct scm_thing * obj = args->car;
  struct scm_thing * alist = args->cdr->car;
  while (alist &&
	 alist->type == PAIR &&
	 alist->car->type == PAIR) {
    if (c_equal_p(obj, alist->car->car)) {
      return alist->car;
    }
    alist = alist->cdr;
  }
  if (alist) {
    printf("assoc run on non-alist\n");
    error();
  }
  return &scm_false;
}

/* 6.3.3. Symbols */

struct scm_thing * scm_symbol_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (args->car && args->car->type == SYMBOL) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_symbol_to_string(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * symbol = args->car;
  if (!symbol || symbol->type != SYMBOL) {
    error();
  } else {
    struct scm_thing * string = malloc(sizeof(struct scm_thing));
    string->type = STRING;
    string->mutable = false;
    string->string = symbol->symbol;
    string->stringsize = strlen(symbol->symbol);
    return string;
  }
}

struct scm_thing * scm_string_to_symbol(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * string = args->car;
  if (!string || string->type != STRING) {
    error();
  } else {
    char * buf = malloc((string->stringsize + 1) * sizeof(char));
    memcpy(buf, string->string, string->stringsize);
    buf[string->stringsize] = '\0';
    struct scm_thing * symbol = malloc(sizeof(struct scm_thing));
    symbol->type = SYMBOL;
    symbol->mutable = false;
    symbol->symbol = buf;
    return symbol;
  }
}

/* 6.3.4. Characters */

static struct scm_thing * create_character(unsigned char character) {
  struct scm_thing * rv = malloc(sizeof(struct scm_thing));
  rv->type = CHARACTER;
  rv->mutable = true;
  rv->character = character;
  return rv;
}

bool c_char_p(struct scm_thing * obj) {
  if (obj && obj->type == CHARACTER) {
    return true;
  } else {
    return false;
  }
}

struct scm_thing * scm_char_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (c_char_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_char_equals_p(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * char1 = args->car;
  struct scm_thing * char2 = args->cdr->car;
  if (!char1 || !char2) {
    error();
  }
  if (char1->type != CHARACTER || char2->type != CHARACTER) {
    error();
  }
  if (char1->character == char2->character) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_char_less_than_p(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * char1 = args->car;
  struct scm_thing * char2 = args->cdr->car;
  if (!char1 || !char2) {
    error();
  }
  if (char1->type != CHARACTER || char2->type != CHARACTER) {
    error();
  }
  if (char1->character < char2->character) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_char_greater_than_p(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * char1 = args->car;
  struct scm_thing * char2 = args->cdr->car;
  if (!char1 || !char2) {
    error();
  }
  if (char1->type != CHARACTER || char2->type != CHARACTER) {
    error();
  }
  if (char1->character > char2->character) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_char_less_than_equals_p(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * char1 = args->car;
  struct scm_thing * char2 = args->cdr->car;
  if (!char1 || !char2) {
    error();
  }
  if (char1->type != CHARACTER || char2->type != CHARACTER) {
    error();
  }
  if (char1->character <= char2->character) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_char_greater_than_equals_p(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * char1 = args->car;
  struct scm_thing * char2 = args->cdr->car;
  if (!char1 || !char2) {
    error();
  }
  if (char1->type != CHARACTER || char2->type != CHARACTER) {
    error();
  }
  if (char1->character >= char2->character) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_char_ci_equals_p(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * char1 = args->car;
  struct scm_thing * char2 = args->cdr->car;
  if (!char1 || !char2) {
    error();
  }
  if (char1->type != CHARACTER || char2->type != CHARACTER) {
    error();
  }
  if (char1->character == char2->character) {
    return &scm_true;
  } else if ((char1->character & 0xdf) == (char2->character & 0xdf) &&
	     (char1->character & 0xdf) >= ('a' & 0xdf) &&
	     (char1->character & 0xdf) <= ('z' & 0xdf)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_char_to_integer(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * character = args->car;
  if (!character || character->type != CHARACTER) {
    error();
  }
  return create_integer(character->character);
}

struct scm_thing * scm_integer_to_char(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * integer = args->car;
  if (!integer || integer->type != REAL_RATIONAL) {
    error();
  }
  if (integer->realdenominator != 1) {
    error();
  }
  if (integer->realnumerator > CHAR_MAX ||
      integer->realnumerator < CHAR_MIN) {
    error();
  }
  return create_character(integer->realnumerator);
}

/* 6.3.5. Strings */

bool valid_index(struct scm_thing * obj) {
  if (obj &&
      obj->type == REAL_RATIONAL &&
      obj->realdenominator == 1 &&
      obj->realnumerator >= 0) {
    return true;
  } else {
    return false;
  }
}

bool c_string_p(struct scm_thing * obj) {
  if (obj && obj->type == STRING) {
    return true;
  } else {
    return false;
  }
}

struct scm_thing * scm_string_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (c_string_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * scm_make_string(struct scm_thing * args) {
  if (c_length(args) < 1 || c_length(args) > 2) {
    error();
  }
  if (!args->car || args->car->type != REAL_RATIONAL) {
    error();
  }
  if (args->car->realdenominator != 1) {
    error();
  }
  if (args->car->realdenominator != 1) {
    error();
  }
  if (c_length(args) == 2) {
    if (!args->cdr->car || args->cdr->car->type != CHARACTER) {
      error();
    }
  }
  struct scm_thing * string = malloc(sizeof(struct scm_thing));
  string->type = STRING;
  string->mutable = true;
  string->stringsize = args->car->realnumerator;
  string->string = malloc(string->stringsize * sizeof(char));
  if (c_length(args) == 2) {
    memset(string->string, args->cdr->car->character, string->stringsize);
  }
  return string;
}

// can't be bothered with string

struct scm_thing * scm_string_length(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (!args->car || args->car->type != STRING) {
    error();
  }
  return create_integer(args->car->stringsize);
}

struct scm_thing * scm_string_ref(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * string = args->car;
  struct scm_thing * k = args->cdr->car;
  if (!string || string->type != STRING) {
    error();
  }
  if (!valid_index(k)) {
    error();
  }
  if (k->realnumerator >= string->stringsize) {
    error();
  }
  return create_character(string->string[k->realnumerator]);
}

struct scm_thing * scm_string_set(struct scm_thing * args) {
  if (c_length(args) != 3) {
    error();
  }
  struct scm_thing * string = args->car;
  struct scm_thing * k = args->cdr->car;
  struct scm_thing * character = args->cdr->cdr->car;
  if (!string || string->type != STRING || !string->mutable) {
    error();
  }
  if (!valid_index(k)) {
    error();
  }
  if (!character || character->type != CHARACTER) {
    error();
  }
  if (k->realnumerator >= string->stringsize) {
    error();
  }
  string->string[k->realnumerator] = character->character;
  return string;
}

// not bothering with the library procedures yet

/* 6.3.6. Vectors */

bool c_vector_p(struct scm_thing * obj) {
  if (obj && obj->type == VECTOR) {
    return true;
  } else {
    return false;
  }
}

struct scm_thing * scm_vector_p(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  if (c_vector_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * c_make_vector(uint32_t k) {
  struct scm_thing * new_vector = malloc(sizeof(struct scm_thing));
  new_vector->type = VECTOR;
  new_vector->vectorsize = k;
  new_vector->vector = malloc(sizeof(struct scm_thing *) * k);
  return new_vector;
}

struct scm_thing * scm_make_vector(struct scm_thing * args) {
  if (c_length(args) == 1) {
    struct scm_thing * k = args->car;
    if (!c_exact_integer_p(k) || k->realnumerator < 0) {
      error();
    }
    return c_make_vector(k->realnumerator);
  } else if (c_length(args) == 2) {
    struct scm_thing * k = args->car;
    struct scm_thing * fill = args->cdr->car;
    if (!valid_index(k)) {
      error();
    }
    struct scm_thing * rv = c_make_vector(k->realnumerator);
    return c_vector_fill(rv, fill);
  } else {
    error();
  }
}
    
struct scm_thing * scm_vector_length(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * vector = args->car;
  if (!c_vector_p(vector)) {
    error();
  }
  return create_integer(vector->vectorsize);
}

struct scm_thing * scm_vector_ref(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * vector = args->car;
  struct scm_thing * k = args->cdr->car;
  if (!c_vector_p(vector)) {
    printf("attempt to call vector-ref with non-vector\n");
    error();
  }
  if (!valid_index(k)) {
    printf("attempt to call vector-ref with invalid index\n");
    error();
  }
  if (k->realnumerator >= vector->vectorsize) {
    printf("attempt to vector-ref with too large of an index\n");
    error();
  }
  return vector->vector[k->realnumerator];
}

struct scm_thing * scm_vector_set(struct scm_thing * args) {
  if (c_length(args) != 3) {
    error();
  }
  struct scm_thing * vector = args->car;
  struct scm_thing * k = args->cdr->car;
  struct scm_thing * obj = args->cdr->cdr->car;
  if (!c_vector_p(vector)) {
    printf("attempt to call vector-set! with non-vector");
    error();
  }
  if (!valid_index(k)) {
    error();
  }
  if (k->realnumerator >= vector->vectorsize) {
    printf("attempt to vector-set! with too large of an index\n");
    error();
  }
  vector->vector[k->realnumerator] = obj;
  return vector;
}

// will program vector->list and list->vector if i actually need them

struct scm_thing * c_vector_fill(struct scm_thing * vector,
				 struct scm_thing * fill) {
  if (vector->vectorsize == 0) {
    return vector;
  } else if (vector->vectorsize == 1) {
    vector->vector[0] = fill;
    return vector;
  } else {
    vector->vector[0] = fill;
    uint32_t i = 1;
    while ((i << 1) < vector->vectorsize) {
      // for some reason internet people use memmove
      // i don't see why---these should not overlap
      // ok i think it might be hypercorrection---people often use memcpy when they should use memmove
      memcpy(vector->vector + i, vector->vector, i * sizeof(struct scm_thing *));
      i = i << 1;
    }
    if (i < vector->vectorsize) {
      memcpy(vector->vector + i, vector->vector, (vector->vectorsize - i) * sizeof(struct scm_thing *));
    }
    return vector;
  }
}

struct scm_thing * scm_vector_fill(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * vector = args->car;
  struct scm_thing * fill = args->cdr->car;
  if (!c_vector_p(vector)) {
    error();
  }
  return c_vector_fill(vector, fill);
}

/* 6.4. Control features */

bool c_procedure_p(struct scm_thing * obj) {
  if (obj &&
      (obj->type == INTERPRETED_PROCEDURE ||
       obj->type == COMPILED_PROCEDURE)) {
    return true;
  } else {
    return false;
  }
}

struct scm_thing * scm_procedure_p(struct scm_thing * args) {
  if (c_procedure_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

// has to be implemented as a macro
struct scm_thing * scm_apply(struct scm_thing * args,
			     struct scm_thing * local_environment,
			     struct scm_thing * top_level_environment) {
  if (c_length(args) < 1) {
    error();
  }
  struct scm_thing * proc = args->car;
  args = args->cdr;
  struct scm_thing * myargs = NULL;
  struct scm_thing * myargs_tail;
  if (args) {
    while (args->cdr) {
      if (!myargs) {
	myargs = malloc(sizeof(struct scm_thing));
	myargs_tail = myargs;
      } else {
	myargs_tail->cdr = malloc(sizeof(struct scm_thing));
	myargs_tail = myargs_tail->cdr;
      }
      myargs_tail->car = args->car;
      args = args->cdr;
    }
    // will break that thing that's supposed to return a newly allocated list
    if (myargs) {
      myargs_tail->cdr = args->car;
    } else {
      myargs = args->car;
    }
  }
  return c_procedure_call(proc,
			  myargs,
			  local_environment,
			  top_level_environment);
}

// not sure whether i'll implement values and call-with-values

/* 6.5. Eval */

struct scm_thing * scm_eval(struct scm_thing * args) {
  if (c_length(args) != 2) {
    error();
  }
  struct scm_thing * expression = args->car;
  struct scm_thing * environment_specifier = args->cdr->car;
  if (environment_specifier && environment_specifier->type != PAIR) {
    error();
  }
  return c_eval(expression,
		NULL,
		environment_specifier);
}

struct scm_thing * scm_null_environment(struct scm_thing * args) {
  return null_variables;
}

/* 6.6. Input and output */

/* 6.6.1. Ports */

bool c_input_port_p(struct scm_thing * obj) {
  if (obj && obj->type == INPUT_PORT) {
    return true;
  } else {
    return false;
  }
}

struct scm_thing * scm_input_port_p(struct scm_thing * args) {
  if (c_input_port_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

bool c_output_port_p(struct scm_thing * obj) {
  if (obj && obj->type == OUTPUT_PORT) {
    return true;
  } else {
    return false;
  }
}

struct scm_thing * scm_output_port_p(struct scm_thing * args) {
  if (c_output_port_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

struct scm_thing * current_input_port;
struct scm_thing * current_output_port;

struct scm_thing * scm_current_input_port(struct scm_thing * args) {
  return current_input_port;
}

struct scm_thing * scm_current_output_port(struct scm_thing * args) {
  return current_output_port;
}

struct scm_thing * scm_open_input_file(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * filename = args->car;
  if (!c_string_p(filename)) {
    error();
  }
  char * buf = malloc((filename->stringsize + 1) * sizeof(char));
  memcpy(buf, filename->string, filename->stringsize);
  buf[filename->stringsize] = '\0';
  struct scm_thing * rv = malloc(sizeof(struct scm_thing));
  rv->type = INPUT_PORT;
  rv->mutable = true;
  rv->portfd = open(buf, O_RDONLY);
  if (rv->portfd < 0) {
    perror("error opening input file");
    error();
  }
  return rv;
}

struct scm_thing * scm_open_output_file(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * filename = args->car;
  if (!c_string_p(filename)) {
    error();
  }
  char * buf = malloc((filename->stringsize + 1) * sizeof(char));
  memcpy(buf, filename->string, filename->stringsize);
  buf[filename->stringsize] = '\0';
  struct scm_thing * rv = malloc(sizeof(struct scm_thing));
  rv->type = OUTPUT_PORT;
  rv->mutable = true;
  rv->portfd = creat(buf, 0662);
  if (rv->portfd < 0) {
    perror("error creating output file");
    error();
  }
  return rv;
}

struct scm_thing * scm_close_input_port(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * port = args->car;
  if (!c_input_port_p(port)) {
    error();
  }
  if (close(port->portfd) < 0) {
    perror("error closing input port");
    error();
  }
  // whatever
  return NULL;
}

struct scm_thing * scm_close_output_port(struct scm_thing * args) {
  if (c_length(args) != 1) {
    error();
  }
  struct scm_thing * port = args->car;
  if (!c_output_port_p(port)) {
    error();
  }
  if (close(port->portfd) < 0) {
    perror("error closing output port");
    error();
  }
  // whatever
  return NULL;
}

/* 6.6.2. Input */

struct scm_thing * scm_read_char(struct scm_thing * args) {
  if (c_length(args) < 0 || c_length(args) > 1) {
    error();
  }
  int fd;
  if (c_length(args) == 0) {
    fd = current_input_port->portfd;
  } else {
    if (!c_input_port_p(args->car)) {
      error();
    }
    fd = args->car->portfd;
  }
  unsigned char character = EOF;
  if (read(fd, &character, 1) < 0) {
    perror("error reading character from input port");
    error();
  }
  return create_character(character);
}

struct scm_thing * scm_peek_char(struct scm_thing * args) {
  if (c_length(args) < 0 || c_length(args) > 1) {
    error();
  }
  int fd;
  if (c_length(args) == 0) {
    fd = current_input_port->portfd;
  } else {
    if (!c_input_port_p(args->car)) {
      error();
    }
    fd = args->car->portfd;
  }
  unsigned char character = EOF;
  ssize_t read_rv = read(fd, &character, 1);
  if (read_rv < 0) {
    perror("error reading character from input port");
    error();
  }
  if (read_rv) {
    if (lseek(fd, -1, SEEK_CUR) < 0) {
      perror("error seeking input port");
      error();
    }
  }
  return create_character(character);
}

bool c_eof_object_p(struct scm_thing * obj) {
  if (c_char_p(obj) &&
      obj->character == (char)EOF) {
    return true;
  } else {
    return false;
  }
}

struct scm_thing * scm_eof_object_p(struct scm_thing * args) {
  if (c_eof_object_p(args->car)) {
    return &scm_true;
  } else {
    return &scm_false;
  }
}

// char-ready? not needed? idk how to implement it

/* 6.6.3. Output */

struct scm_thing * scm_write_char(struct scm_thing * args) {
  if (c_length(args) < 1 || c_length(args) > 2) {
    error();
  }
  struct scm_thing * character = args->car;
  if (!c_char_p(character)) {
    error();
  }
  int fd;
  if (c_length(args) == 1) {
    fd = current_output_port->portfd;
  } else {
    if (!c_output_port_p(args->cdr->car)) {
      error();
    }
    fd = args->cdr->car->portfd;
  }
  if (write(fd, &character->character, 1) < 0) {
    perror("error writing character to output port");
    error();
  }
  return NULL;
}

/* 6.6.4. System interface */
  
