(load "macros.scm")
(load "library.scm")

(define (vector-slice oldvector k n)
  (letrec ((newvector (make-vector n))
	   (internal-loop
	    (lambda (i)
	      (vector-set! newvector
			   i
			   (vector-ref oldvector (+ k i)))
	      (if (< (+ i 1) n)
		  (internal-loop (+ i 1))
		  newvector))))
    (internal-loop 0)))

(define (make-vector-object vector)
  (let ((point 0))
    (lambda (operation arg)
      (cond ((equal? operation 'seek-abs)
	     (set! point arg))
	    ((equal? operation 'seek-rel)
	     (set! point (+ point arg)))
	    ((equal? operation 'read)
	     (set! point (+ arg point))
	     (vector-slice vector
			   (- point arg)
			   arg))
	    ((equal? operation 'get-point)
	     point)))))

(define (assq-ref alist key)
  (let ((pair (assq key alist)))
    (if pair
	(cdr pair)
	#f)))

(define (match-vobj-alist vobj default-size default-conversions symbols alist)
  (if (null? symbols)
      alist
      (let* ((symbol (car symbols))
	     (name
	      (if (pair? symbol)
		  (car symbol)
		  symbol))
	     (size
	      (if (pair? symbol)
		  (car (cdr symbol))
		  default-size))
	     (conversion
	      (cond ((and (pair? symbol) (not (null? (cdr (cdr symbol)))))
		     (car (cdr (cdr symbol))))
		    ((assq-ref default-conversions size))
		    (else identity)))
	     (value
	      (conversion (vobj 'read size))))
	(if name
	    (match-vobj-alist vobj
			      default-size
			      default-conversions
			      (cdr symbols)
			      (cons (cons name value)
				    alist))
	    (match-vobj-alist vobj
			      default-size
			      default-conversions
			      (cdr symbols)
			      alist)))))

(define (read-elf-header vobj)
  (match-vobj-alist
   vobj
   1
   uint-conversions
   '(ei-mag0
     ei-mag1
     ei-mag2
     ei-mag3
     ei-class
     ei-data
     ei-version
     ei-osabi
     ei-abiversion
     #f
     #f
     #f
     #f
     #f
     #f
     #f
     (e-type 2)
     (e-machine 2)
     (e-version 4)
     (e-entry 4)
     (e-phoff 4)
     (e-shoff 4)
     (e-flags 4)
     (e-ehsize 2)
     (e-phentsize 2)
     (e-phnum 2)
     (e-shentsize 2)
     (e-shnum 2)
     (e-shstrndx 2))
   '()))

(define (read-section-header vobj)
  (match-vobj-alist
   vobj
   4
   uint-conversions
   '(sh-name
     sh-type
     sh-flags
     sh-addr
     sh-offset
     sh-size
     sh-link
     sh-info
     sh-addralign
     sh-entsize)
   '()))

(define (read-symbol-entry vobj)
  (match-vobj-alist
   vobj
   4
   uint-conversions
   (list 'st-name
	 'st-value
	 'st-size
	 '(st-info 1)
	 '(st-other 1)
	 '(st-shndx 2))
   '()))

(define (read-rel-entry vobj)
  (match-vobj-alist
   vobj
   4
   uint-conversions
   '(r-offset
     r-info)
   '()))

(define (read-rela-entry vobj)
  (match-vobj-alist
   vobj
   4
   uint-conversions
   '(r-offset
     r-info
     r-addend)
   '()))

(define (read-section-headers vobj elf-header)
  (vobj 'seek-abs (assq-ref elf-header 'e-shoff))
  (let* ((shnum (cdr (assoc 'e-shnum elf-header)))
	 (headers
	  (do ((i 0 (+ i 1))
	       (vector (make-vector shnum)))
	      ((= i shnum) vector)
	    (vector-set! vector
			 i
			 (read-section-header vobj)))))
    (get-names vobj
	       headers
	       'sh-name
	       (assq-ref (vector-ref headers
				     (assq-ref elf-header 'e-shstrndx))
			 'sh-offset))))

;; section headers must have names as strings
(define (read-symbol-table vobj section-headers symtab)
  (let ((entries (/ (assq-ref symtab 'sh-size)
		    (assq-ref symtab 'sh-entsize)))
	(strtab (vector-ref section-headers
			    (assq-ref symtab 'sh-link))))
    (vobj 'seek-abs (assq-ref symtab 'sh-offset))
    (get-names vobj
	       (do ((i 0 (+ i 1))
		    (vector (make-vector entries)))
		   ((= i entries) vector)
		 (vector-set! vector i (read-symbol-entry vobj)))
	       'st-name
	       (assq-ref strtab 'sh-offset))))

(define (read-rel-table vobj rel-sh)
  (let ((entries (/ (assq-ref rel-sh 'sh-size)
		    (assq-ref rel-sh 'sh-entsize))))
    (vobj 'seek-abs (assq-ref rel-sh 'sh-offset))
    (do ((vector (make-vector entries))
	 (i 0 (+ i 1)))
	((= i entries) vector)
      (vector-set! vector i (read-rel-entry vobj)))))

(define (read-rela-table vobj rela-sh)
  (let ((entries (/ (assq-ref rela-sh 'sh-size)
		    (assq-ref rela-sh 'sh-entsize))))
    (vobj 'seek-abs (assq-ref rela-sh 'sh-offset))
    (do ((vector (make-vector entries))
	 (i 0 (+ i 1)))
	((= i entries) vector)
      (vector-set! vector i (read-rela-entry vobj)))))

(define (fill-out-symbol-table section-headers symbol-table)
  (do ((i 0 (+ i 1)))
      ((= i (vector-length symbol-table))
       symbol-table)
    (let ((shndx (assq 'st-shndx (vector-ref symbol-table i))))
      (set-cdr! shndx
		(if (equal? (cdr shndx) 0)
		    #f
		    (assq-ref (vector-ref section-headers
					  (cdr shndx))
			      'sh-name))))))

(define (find-entry table symbol value)
  (do ((i 0 (+ i 1))
       (last (- (vector-length table) 1)))
      ((or (= i last)
	   (equal? (assq-ref (vector-ref table i) symbol) value))
       (vector-ref table i))))

(define (charvector->u16 charvector)
  (+ (char->integer (vector-ref charvector 0))
     (* 256 (char->integer (vector-ref charvector 1)))))

(define (charvector->u32 charvector)
  (+ (char->integer (vector-ref charvector 0))
     (* 256 (char->integer (vector-ref charvector 1)))
     (* 65536 (char->integer (vector-ref charvector 2)))
     (* 16777216 (char->integer (vector-ref charvector 3)))))

(define (read-string vobj list)
  (let ((char (vector-ref (vobj 'read 1) 0)))
    (if (char=? char (integer->char 0))
	(list->string (reverse list))
	(read-string vobj (cons char list)))))

(define (save-restore-point vobj thunk)
  (let* ((mark (vobj 'get-point 0))
	 (result (thunk)))
    (vobj 'seek-abs mark)
    result))

(define (retrieve-string offset vobj)
  (save-restore-point
   vobj
   (lambda ()
     (vobj 'seek-abs offset)
     (read-string vobj '()))))

(define uint-conversions
  (list
   (cons 1 (lambda (charvector)
	     (char->integer (vector-ref charvector 0))))
   (cons 2 charvector->u16)
   (cons 4 charvector->u32)))

(define (get-names vobj alists symbol offset)
  (do ((i 0 (+ i 1)))
      ((= i (vector-length alists)) alists)
    (set-cdr! (assq symbol (vector-ref alists i))
	      (let ((index (assq-ref (vector-ref alists i) symbol)))
		(if (= index 0)
		    #f
		    (retrieve-string (+ offset index) vobj))))))

(define (filter-alist alist identifiers filtered)
  (if (null? identifiers)
      filtered
      (filter-alist alist
		    (cdr identifiers)
		    (cons (assq (car identifiers) alist)
			  filtered))))

(define (filter-alists alists identifiers test)
  (do ((i 0 (+ i 1))
       (filtered '()))
      ((= i (vector-length alists)) filtered)
    (if (test (vector-ref alists i))
	(set! filtered
	      (cons (filter-alist (vector-ref alists i)
				  identifiers
				  '())
		    filtered)))))

(define (add-entry-to-alist-in-vector vector k identifier obj)
  (vector-set! vector
	       k
	       (cons (cons identifier obj)
		     (vector-ref vector k))))

(define (fill-out-section-headers vobj section-headers)
  (do ((i 0 (+ i 1)))
      ((= i (vector-length section-headers)) section-headers)
    (let* ((section-header (vector-ref section-headers i))
	   (type (assq-ref section-header 'sh-type))
	   (link (assq 'sh-link section-header))
	   (info (assq 'sh-info section-header)))
      (add-entry-to-alist-in-vector
       section-headers
       i
       'contents
       (cond ((or (= type 2) (= type 11)) ;; SHT_SYMTAB or SHT_DYNSYM
	      (read-symbol-table vobj
				 section-headers
				 section-header))
	     ((= type 4)
	      (read-rela-table vobj section-header)) ;; SHT_RELA
	     ((= type 9)
	      (read-rel-table vobj section-header)) ;; SHT_REL
	     ((= (assq-ref section-header 'sh-size) 0) #f)
	     (else
	      (begin (vobj 'seek-abs (assq-ref section-header 'sh-offset))
		     (vobj 'read (assq-ref section-header 'sh-size))))))
      ;; (set-cdr! link
      ;; 		(if (= (cdr link) 0)
      ;; 		    #f
      ;; 		    (assq-ref (vector-ref section-headers
      ;; 					  (cdr link))
      ;; 			      'sh-name)))
      ;; (set-cdr! info
      ;; 		(cond ((or (= type 4) (= type 9)) ;; SHT_RELA or SHT_REL
      ;; 		       (assq-ref (vector-ref section-headers
      ;; 					     (cdr info))
      ;; 				 'sh-name))
      ;; 		      ((= (cdr info) 0) #f)
      ;; 		      (else (cdr info))))
		      )))

(define (get-data-sections section-headers)
  (do ((data-sections '())
       (i 0 (+ i 1)))
      ((= i (vector-length section-headers))
       data-sections)
    (let* ((section-header (vector-ref section-headers i))
	   (type (assq-ref section-header 'sh-type)))
      (if (or (= type 1)           ;; SHT_PROGBITS
	      (= type 8)           ;; SHT_NOBITS
	      (= type #b01110000000000000000000000000011)) ;; SHT_ARM_ATTRIBUTES
	  (set! data-sections
		(cons (cons i section-header)
		      data-sections))))))

(define (add-symbols-to-data-sections symbols stb-local data-sections)
  (do ((i 1 (+ i 1)))
      ((= i stb-local))
    (let* ((symbol (vector-ref symbols i))
	   (section-pair (assq (assq-ref symbol 'st-shndx)
			       data-sections))
	   (local-symbols-pair (assq 'local-symbols
				     (cdr section-pair))))
      (if local-symbols-pair
	  (set-cdr! local-symbols-pair
		    (cons symbol (cdr local-symbols-pair)))
	  (set-cdr! section-pair
		    (cons (cons 'local-symbols
				(list symbol))
			  (cdr section-pair))))))
  (do ((i stb-local (+ i 1)))
      ((= i (vector-length symbols)))
    (let* ((symbol (vector-ref symbols i))
	   (section-pair (assq (assq-ref symbol 'st-shndx)
			       data-sections))
	   (global-symbols-pair (assq 'global-symbols
				      (cdr section-pair))))
      (if global-symbols-pair
	  (set-cdr! global-symbols-pair
		    (cons symbol (cdr global-symbols-pair)))
	  (set-cdr! section-pair
		    (cons (cons 'global-symbols
				(list symbol))
			  (cdr section-pair)))))))

(define (add-rels-to-data-section rels target-pair symbols)
  (if (not (assq 'rels (cdr target-pair)))
      (set-cdr! target-pair
		(cons (cons 'rels '())
		      (cdr target-pair))))
  (do ((i 0 (+ i 1))
       (rels-pair (assq 'rels (cdr target-pair))))
      ((= i (vector-length rels)))
    (let* ((rel (vector-ref rels i))
	   (r-offset (assq-ref rel 'r-offset))
	   (r-info (assq-ref rel 'r-info))
	   (r-sym (quotient r-info 256))
	   (r-type (remainder r-info 256)))
      (set-cdr! rels-pair
		(cons (list (cons 'r-offset r-offset)
			    (cons 'r-type r-type)
			    (cons 'r-sym r-sym)
			    (cons 'symbol (vector-ref symbols r-sym)))
		      (cdr rels-pair))))
    )
  )

;; next to add to this procedure: rela
(define (consolidate-sections section-headers)
  (let ((data-sections
	 (get-data-sections section-headers)))
    (do ((i 0 (+ i 1)))
	((= i (vector-length section-headers))
	 data-sections)
      (let* ((section-header (vector-ref section-headers i))
	     (type (assq-ref section-header 'sh-type)))
	(cond ((= type 2)
	       (add-symbols-to-data-sections
		(assq-ref section-header 'contents)
		(assq-ref section-header 'sh-info)
		data-sections))
	      ((= type 9)
	       (add-rels-to-data-section
		(assq-ref section-header 'contents)
		(assq (assq-ref section-header 'sh-info)
		      data-sections)
		(assq-ref
		 (vector-ref section-headers
			     (assq-ref section-header 'sh-link))
		 'contents)))))
      )))

(let* ((object-file
	(make-vector-object (snarf-file "hello-world.o")))
       (elf-header
	(read-elf-header object-file))
       (section-headers
	(fill-out-section-headers
	 object-file
	 (read-section-headers object-file elf-header)))
       )
  (list
   elf-header
   ;;section-headers
   (consolidate-sections section-headers)
   ))

